#!/usr/bin/perl
BEGIN {
	unshift(@INC,'/etc/aosi');
}
use config;
use Config::General;
use strict;
use warnings;
use utf8;
use JSON::XS;
use LWP::UserAgent;
use Data::Dumper;
use HTTP::Request;
use Net::LDAP::LDIF;
use File::Copy;
use IO::Compress::Gzip qw(gzip $GzipError);
BEGIN {
	unshift(@INC,$config::lib_dir);
}
use Plugins::o365connectAttributes;
use Plugins::LDIFManip;
use List::Util qw(shuffle);

my $VERSION  = 1.0;
my $plugin_name = 'o365connect';

my $conf = new Config::General(
    -ConfigFile      => $config::plugins_dir . '/'.$plugin_name.'.conf',
    -InterPolateVars => 1,
    -AutoLaunder     => 1
);
my %cfg = $conf->getall;

my $base_dn=$config::base_dn;
my $realm =Plugins::LDIFManip::base2realm($base_dn);
my $clientID=$cfg{'ofc_clientID'};
my $clientSecret=$cfg{'ofc_clientSecret'};
my $ldif_file=$config::log_dir . '/' . $cfg{'ofc_file_prefix'} ."_". $realm . $cfg{'ofc_file_ext'};
my $archive_processed_ldif=$cfg{'ofc_archive_processed_ldif'};
#koliko sekundi plugin ceka za dodjelu licenci nakon dodavanja korisnika
our $pauza_s=1;

my %attr_map=%{$Plugins::o365connectAttributes::attr_map};
my @upd_attr=[$Plugins::o365connectAttributes::upd_attr];
my %group_map=%{$Plugins::o365connectAttributes::group_map};
my %licence_map=%{$Plugins::o365connectAttributes::licence_map};

my $userAgent='AAI@EduHrO365Connect/' . "$VERSION";
# Endpoint na kojem se trazi autentikacijski token
my $OauthTokenEndpoint = 'https://login.microsoftonline.com/' . $realm . '.hr/oauth2/token?api-version=1.0';

# GRAPH API URL - resurs za kojeg se trazi token
my $resAzureGraphAPI = "https://graph.windows.net";
 
# Endpoint za administraciju imenika putem graph API-a
my $serviceRootURL = $resAzureGraphAPI . '/' . $realm . '.hr/';

my $graphApiVersion="api-version=1.6";

my $authToken;
my $false=JSON::XS::false;
my $true=JSON::XS::true;
$Data::Dumper::Terse = 1;  
my $ua = LWP::UserAgent->new;
$ua->agent($userAgent);

my %licences;
my %groups;


if (-e $ldif_file) {
	print my_time() . "\n";
	my $new_file=$ldif_file . "." . timestamp();
	move($ldif_file, $new_file);
	my $ldif = Net::LDAP::LDIF->new( "$new_file", "r", onerror => 'die');
	while (my $entry = $ldif->read_entry())
	{
		my $ctype=$entry->changetype();
		my $op_res;
		print $entry->dn() . ": $ctype: ";
		if ($ctype eq 'add') {
			if($op_res=add_user($entry)){
				print "Korisnik dodan u Azure AD. ";
				#pricekati prije dodavanja licence
				sleep($pauza_s);
				add_licence($entry);
				add_user_to_group($entry,$op_res);
			} else {
				print "Greska!";
			}
		} elsif ($ctype eq 'delete'){
			if(delete_user($entry->dn())){
				print "OK";
			}
		} elsif ($ctype eq 'modify'){
			if(modify_user($entry)){
				print "OK";
			}
		} else {
			print "Nedefiniran tip promjene. Preskacem.";
		}
		print "\n";
	}
	$ldif->done();
	
	if($archive_processed_ldif){
		gzip $new_file => "$new_file.gz" or die "gzip failed: $GzipError\n";
	}
	unlink $new_file or warn "Could not unlink $new_file: $!";
	$authToken='';
	%licences=();
	%groups=();
}

sub add_user{
	my $entry=shift;
	my %user=(
		'usagelocation'=>'HR',
		'accountEnabled'=>$true,
		'passwordProfile'=>{'password'=>pwd_gen(),'forceChangePasswordNextLogin'=> $false},
	);
	foreach my $key (keys %attr_map){
		$user{$attr_map{$key}}=$entry->get_value($key);
	}
	return call_azure_method('POST','users',201,\%user);	
}

sub get_user{
	my $dn=shift;
	my $uid=find_uid($dn);
	if($uid) {
		return call_azure_method('GET','users/' . $uid .'@' . $realm . '.hr',200);
	} else {
		error_log("get_user: nije pronadjen uid za dn:" . $dn,0);
	}
}

sub add_licence{
	my $entry=shift;
	my $uid=find_uid($entry->dn());
	my $affiliation=makni_hrvatske($entry->get_value('hrEduPersonPrimaryAffiliation'));
	#koliko puta ponovo pokusa dodati licencu
	my $retries=3;
	
	my @lic;
	if($licence_map{$affiliation}){
		@lic=@{$licence_map{$affiliation}};
	}
	if($uid) {
		if(scalar @lic>0){
			get_licences() unless %licences;
			error_log("Greska u dohvatu podataka o licencama.",1) unless %licences;
			
			while(my $skuPartNumber=shift @lic){
				if($licences{$skuPartNumber}{'capabilityStatus'} eq 'Enabled' && $licences{$skuPartNumber}{'leftUnits'}>0){
					my $params={
	          'addLicenses' => [
											    {
											      "disabledPlans"=> [],
											      "skuId"=> $licences{$skuPartNumber}{'skuId'}
											    }],
	          'removeLicenses' => []
	        };
	        my $i=0;
	        while($i < $retries ){
				    if(call_azure_method('POST','users/' . $uid .'@' . $realm . '.hr/assignLicense',200,$params)){
				    		$licences{$skuPartNumber}{'leftUnits'}--;
				    		$licences{$skuPartNumber}{'consumedUnits'}++;
				    		print "Dodijeljena licenca $skuPartNumber. ";
				    		return 1;
				    }
				    print "\nDodjela licence nije uspjela. Pokusavam ponovo.\n";
				    sleep($pauza_s);
				    $i++;
				 }
				} else {
					error_log("add_licence: Nema preostalih omogucenih licenci tipa $skuPartNumber. capabilityStatus: " . $licences{$skuPartNumber}{'capabilityStatus'} . ", leftUnits:" . $licences{$skuPartNumber}{'leftUnits'},0);
				}
			}
		} else {
			print "add_licence: Za hrEduPersonPrimaryAffiliation '$affiliation' nije podesena defaultna licenca.";
		}
	} else {
		error_log("add_licence: nije pronadjen uid za dn:" . $entry->dn(),0);
	}
}

sub add_user_to_group{
	my $entry=shift;
	my $ADentry=shift;
	my $uid=find_uid($entry->dn());
	my $objectId=$ADentry->{'objectId'};
	my $affiliation=makni_hrvatske($entry->get_value('hrEduPersonPrimaryAffiliation'));
	#koliko puta ponovo pokusa dodati grupu
	my $retries=3;
	my @group;
	if($group_map{$affiliation}){
		@group=@{$group_map{$affiliation}};
	}
	if($uid) {
		if(scalar @group>0){
			get_groups() unless %groups;
			while(my $gr=shift @group){
				create_group($gr) unless ($groups{$gr});
				if($groups{$gr}){
					my $p={
					  "url"=>"https://graph.windows.net/" . $realm . ".hr/directoryObjects/$objectId"
					};
					my $i=0;
	        	while($i < $retries ){
						if(my $result=call_azure_method('POST','groups/' . $groups{$gr} . '/$links/members',204,$p)){
							print "Korisnik dodan u grupu $gr.";
							return $result;
						}
						print "\nNije uspjelo dodavanje u grupu. Pokusavam ponovo.\n";
						sleep($pauza_s);
						$i++;
					}
				} else {
					print "add_user_to_group: Za hrEduPersonPrimaryAffiliation '$affiliation' ne postoji grupa $gr u Azure AD.";
				}
			}
		} else {
			print "add_user_to_group: Za hrEduPersonPrimaryAffiliation '$affiliation' nije podesena defaultna grupa.";
		}
	} else {
		error_log("add_user_to_group: nije pronadjen uid za dn:" . $entry->dn(),0);
	}
}

sub delete_user{
	my $dn=shift;
	my $uid=find_uid($dn);
	if($uid) {
		return call_azure_method('DELETE','users/' . $uid .'@' . $realm . '.hr',204);
	} else {
		error_log("delete_user: nije pronadjen uid za dn:" . $dn,0);
	}
}

sub modify_user{
	my $entry=shift;
	my %user;
	my $uid=find_uid($entry->dn());
	if($uid) {
		my @changes=$entry->changes;
		while(my $cst=shift @changes){
			my @a=shift @changes;
			if($cst eq 'replace'){
				$user{$attr_map{$a[0][0]}}=$a[0][1][0];
			}
		}
		my $size = scalar keys %user;
		if($size>0){
			return call_azure_method('PATCH','users/' . $uid .'@' . $realm . '.hr',204,\%user);
		} else {
			return 1;
		}
	} else {
		error_log("modify_user: nije pronadjen uid za dn:" . $entry->dn(),0);
	}
}

sub get_licences{
	my $l=call_azure_method('GET','subscribedSkus',200);
	my @lic=@{$l->{'value'}};
	while (my $licence=shift @lic){
		$licences{$licence->{'skuPartNumber'}}{'skuId'}=$licence->{'skuId'};
		$licences{$licence->{'skuPartNumber'}}{'capabilityStatus'}=$licence->{'capabilityStatus'};
		$licences{$licence->{'skuPartNumber'}}{'consumedUnits'}=$licence->{'consumedUnits'};
		$licences{$licence->{'skuPartNumber'}}{'prepaidUnits'}=$licence->{'prepaidUnits'}->{'enabled'};
		$licences{$licence->{'skuPartNumber'}}{'leftUnits'}=$licence->{'prepaidUnits'}->{'enabled'} - $licence->{'consumedUnits'};
	}
}

sub create_group{
	my $displayName=shift;
	my %group=(
		'displayName'=>$displayName,
		'mailEnabled'=>$false,
		'mailNickname'=>$displayName,
		'securityEnabled'=>$true
	);
	if(call_azure_method('POST','groups',201,\%group)){
		get_groups();
	}
}


sub get_groups{
	my $g=call_azure_method('GET','groups',200);
	my @gr=@{$g->{'value'}};
	while (my $group=shift @gr){
		$groups{$group->{'displayName'}}=$group->{'objectId'};
	}
}

sub find_uid{
	my $dn=shift;
	if($dn =~ m/uid=([^\s,]+)[\s,]/) {
		return $1;
	}
}

sub set_auth_header{
	my $reqRef = shift;
	get_auth_token() unless($authToken->{'expires_on'} && $authToken->{'expires_on'}>time);
	$$reqRef->header('Authorization' => $authToken->{'token_type'} . ' ' . $authToken->{'access_token'}, 'Content-Type'=>'application/json; charset=utf-8', 'Content-Transfer-Encoding'=>'binary');
}

sub get_auth_token{
	my $response=$ua->post($OauthTokenEndpoint, [
		'grant_type'=>'client_credentials',
		'client_secret'=>$clientSecret,
		'resource'=> $resAzureGraphAPI,
		'client_id' => $clientID
		]
	);
	error_log("get_auth_token:\nstatus:" . $response->status_line . "\nresponse:"  . $response->content, 1)
		unless $response->is_success;
	$authToken  = decode_json($response->content);
}

sub call_azure_method{
	my $http_method=shift;
	my $method=shift;
	my $code_ok=shift;
	my $hash_content_ref=shift;
	
	my $content;
	
	my $req = HTTP::Request->new( $http_method, $serviceRootURL . $method . '?' . $graphApiVersion );
	set_auth_header(\$req);
	if($hash_content_ref){
		$content=JSON::XS->new->utf8(0)->encode($hash_content_ref);
		$req->content($content);
	}
	
	my $response=$ua->request($req);
	if($response->code==$code_ok){
		if(length($response->content)>0){
			return decode_json($response->content);
		} else {
			return 1;
		}
	} else {
		error_log("call_azure_method: $http_method, $method\ncontent:" . Dumper($content) . "status:" . $response->status_line . "\nresponse:"  . $response->content, 0);
		return 0;
	}
}


sub error_log{
	my $message=shift;
	my $umri=shift;
	
	#print STDERR my_time() . " " . $message;
	print $message;
	
	if($umri==1){
		#print STDERR "exiting\n";
		print " izlazim\n";
		exit;
	}
	
}

sub makni_hrvatske{
	my $txt=shift;
	$txt =~ s/[^[:ascii:]]//g;
	return $txt;
}

sub dump_content{
	my $content=shift;
	return Dumper(decode_json($content));
}

sub pwd_gen{
	my @a = split '', 'qwertyuiopasdfghjklzxcvbnm';
	my @b = split '', 'QWERTYUIOPASDFGHJKLZXCVBNM';
	my @c = split '', ',.?<>;:@/!%^&*()-+=_';
	my @d = split '', '0123456789';
	my @e = split '', 'qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM,.?<>;:@/!%^&*()-+=_0123456789';
	my $pass=pick_chrs(\@a,2);
	$pass.=pick_chrs(\@b,2);
	$pass.=pick_chrs(\@c,2);
	$pass.=pick_chrs(\@d,2);
	$pass.=pick_chrs(\@e,2);
	return join('',shuffle(split('',$pass)));
}
sub pick_chrs{
	my $c_ref=shift;
	my $num=shift;
	my @c=@{$c_ref};
	my $pass='';
	for(my $i=0;$i<$num;$i++){
		$pass .= $c[rand @c];
	}
	return $pass;
}

sub my_time {
	my ($second, $minute, $hour, $dayOfMonth, $month, $yearOffset, $dayOfWeek, $dayOfYear, $daylightSavings) = localtime();
	my $year = 1900 + $yearOffset;
	$month++;
	return "$year-$month-$dayOfMonth $hour:$minute:$second";
}

sub timestamp {
	my ($second, $minute, $hour, $dayOfMonth, $month, $yearOffset, $dayOfWeek, $dayOfYear, $daylightSavings) = localtime();
	my $year = 1900 + $yearOffset;
	$month++;
	return "$year-$month-$dayOfMonth-$hour-$minute-$second";
}